--- old/usr/src/lib/libzfs/common/libzfs_sendrecv.c	2013-04-25 12:27:56.101884635 +0000
+++ new/usr/src/lib/libzfs/common/libzfs_sendrecv.c	2013-04-25 12:27:56.108107276 +0000
@@ -793,6 +793,7 @@
 	int outfd;
 	boolean_t err;
 	nvlist_t *fss;
+	nvlist_t *snapholds;
 	avl_tree_t *fsavl;
 	snapfilter_cb_t *filter_cb;
 	void *filter_cb_arg;
@@ -951,14 +952,12 @@
 
 	assert(zhp->zfs_type == ZFS_TYPE_SNAPSHOT);
 
-	if (sdd->dryrun)
-		return (0);
-
 	/*
-	 * zfs_send() only opens a cleanup_fd for sends that need it,
-	 * e.g. replication and doall.
+	 * We process if snapholds is not NULL even if on a dry run as
+	 * this is used to pre-calculate the required holds so they can
+	 * be processed in one kernel request
 	 */
-	if (sdd->cleanup_fd == -1)
+	if (sdd->snapholds == NULL)
 		return (0);
 
 	thissnap = strchr(zhp->zfs_name, '@') + 1;
@@ -971,8 +970,8 @@
 	 * handle that error.
 	 */
 	if (pzhp) {
-		error = zfs_hold(pzhp, thissnap, sdd->holdtag,
-		    B_FALSE, B_TRUE, sdd->cleanup_fd);
+		error = zfs_hold_add(pzhp, thissnap, sdd->holdtag, B_TRUE,
+		    sdd->snapholds);
 		zfs_close(pzhp);
 	}
 
@@ -1542,8 +1541,10 @@
 			err = errno;
 			goto stderr_out;
 		}
+		sdd.snapholds = fnvlist_alloc();
 	} else {
 		sdd.cleanup_fd = -1;
+		sdd.snapholds = NULL;
 	}
 	if (flags->verbose) {
 		/*
@@ -1565,6 +1566,30 @@
 			    "total estimated size is %s\n"), buf);
 		}
 	}
+
+	if (sdd.snapholds != NULL) {
+		/* Holds are required */
+		if (!flags->verbose) {
+			/*
+			 * A verbose dry run wasn't done so do a non-verbose
+			 * dry run to collate snapshot hold's.
+			 */
+			sdd.dryrun = B_TRUE;
+			err = dump_filesystems(zhp, &sdd);
+			sdd.dryrun = flags->dryrun;
+		}
+
+		if (err != 0) {
+			fnvlist_free(sdd.snapholds);
+			goto stderr_out;
+		}
+
+		err = zfs_hold_apply(zhp, B_TRUE, sdd.cleanup_fd, sdd.snapholds);
+		fnvlist_free(sdd.snapholds);
+		if (err != 0)
+			goto stderr_out;
+	}
+	
 	err = dump_filesystems(zhp, &sdd);
 	fsavl_destroy(fsavl);
 	nvlist_free(fss);
